@@ -39,7 +39,7 @@ T = TypeVar('T')
 
 def tree_map_rngs(fn, tree):
   """Needed for mapping JAX random.* functions over KeyArray leaves."""
-  return jax.tree_map(
+  return jax.tree_util.tree_map(
       fn, tree, is_leaf=lambda x: isinstance(x, random.KeyArray))
 
 
@@ -109,7 +109,7 @@ def pack(fn: Callable[..., Any],
       msg = 'kwargs are not supported in {}, so \"{}\" is(are) ignored'
       warnings.warn(msg.format(name, ', '.join(kwargs.keys())), RuntimeWarning)
     # pylint: disable=protected-access
-    scopes, treedef = jax.tree_flatten(scope_tree)
+    scopes, treedef = jax.tree_util.tree_flatten(scope_tree)
     scopes, paths = _dedup_scopes(scopes)
 
     variable_groups_xs = []
@@ -169,7 +169,7 @@ def pack(fn: Callable[..., Any],
           rngs.update(rng_group)
         # make sure variable dicts are cloned and can't be manipulated by ref
         # sharing.
-        variables = jax.tree_map(lambda x: x, variables)
+        variables = jax.tree_util.tree_map(lambda x: x, variables)
         scope_mutable = intersect_filters(
             intersect_filters(scope.mutable, mutable), mutable_filter)
         new_path = scope.path
@@ -181,7 +181,7 @@ def pack(fn: Callable[..., Any],
         inner_scope = Scope(
             variables, name=scope.name, rngs=rngs,
             mutable=scope_mutable, parent=None,
-            path=new_path)
+            path=new_path, flags=scope.flags)
         inner_scope.rng_counters = rng_counters
         inner_scopes.append(inner_scope)
       inner_scopes = _dup_scopes(scopes, inner_scopes, paths)
@@ -271,7 +271,7 @@ def map_variables(fn: Callable[..., Any],
     if init:
       scopes = scope_fn((target, variables), rng_groups)
       has_mutable_cols = any(not is_filter_empty(scope.mutable)
-                             for scope in jax.tree_leaves(scopes))
+                             for scope in jax.tree_util.tree_leaves(scopes))
       if has_mutable_cols:
         fn(scopes, *args, **kwargs)
         target, _ = repack(scopes)
@@ -421,7 +421,7 @@ def vjp(
     y, bwd, (aux, out_vars) = jax.vjp(
         wrapper, vjp_vars, *args,
         reduce_axes=reduce_axes, has_aux=True)
-    treedef = jax.tree_structure(scope)
+    treedef = jax.tree_util.tree_structure(scope)
     bwd = jax.tree_util.Partial(
         functools.partial(_bwd_wrapper, treedef), bwd)
     if has_aux:
@@ -458,7 +458,8 @@ def jvp(
       return p * x
 
     def f(scope, x):
-      vars_t = jax.tree_map(jnp.ones_like, scope.variables().get('params', {}))
+      vars_t = jax.tree_util.tree_map(jnp.ones_like,
+                                      scope.variables().get('params', {}))
       x, out_t = lift.jvp(
           learn_scale, scope, (x,), (jnp.zeros_like(x),),
           variable_tangents={'params': vars_t})
@@ -503,7 +504,7 @@ def jvp(
     return (y, out_tangents[0]), out_vars
   # filter out empty tangent collections because JAX will error on non-equal
   # tree structure for example: {"params": {}} != {}.
-  treedef = jax.tree_structure(scope)
+  treedef = jax.tree_util.tree_structure(scope)
 
   variable_tangents = tuple({k: v  # pylint: disable=g-complex-comprehension
                              for k, v in vt.items()
@@ -518,9 +519,11 @@ def jvp(
 def vmap(fn: Callable[..., Any],
          variable_axes: Mapping[CollectionFilter, InOutAxis],
          split_rngs: Mapping[PRNGSequenceFilter, bool],
-         in_axes=0, out_axes=0,
+         in_axes=0,
+         out_axes=0,
          axis_size: Optional[int] = None,
-         axis_name: Optional[str] = None) -> Callable[..., Any]:
+         axis_name: Optional[str] = None,
+         spmd_axis_name: Optional[str] = None) -> Callable[..., Any]:
   """A lifted version of ``jax.vmap``.
 
   See ``jax.vmap`` for the unlifted batch transform in Jax.
@@ -568,6 +571,9 @@ def vmap(fn: Callable[..., Any],
     axis_name: Specifies a name for the batch axis. Can be used together
       with parallel reduction primitives (e.g. `jax.lax.pmean`,
       `jax.lax.ppermute`, etc.)
+    spmd_axis_name: Axis name added to any pjit sharding constraints appearing
+      in `fn`. See also
+      https://github.com/google/flax/blob/main/flax/linen/partitioning.py.
 
   Returns:
     A vectorized version of the input scope function.
@@ -581,15 +587,16 @@ def vmap(fn: Callable[..., Any],
   def inner(scope_fn, repack_fn, variable_groups, rng_groups, *args):
     def find_axis_size(axis, x):
       if axis is not None:
-        leaves = jax.tree_leaves(x)
+        leaves = jax.tree_util.tree_leaves(x)
         if leaves:
           return leaves[0].shape[axis]
       return ()
 
     # split rngs
-    axis_sizes = jax.tree_map(find_axis_size, (variable_in_axes, in_axes),
-                                   (variable_groups, args))
-    axis_sizes = set(jax.tree_leaves(axis_sizes))
+    axis_sizes = jax.tree_util.tree_map(find_axis_size,
+                                        (variable_in_axes, in_axes),
+                                        (variable_groups, args))
+    axis_sizes = set(jax.tree_util.tree_leaves(axis_sizes))
     if axis_size is None and len(axis_sizes) == 1:
       d_axis_size, = axis_sizes
     elif len(axis_sizes) > 1:
@@ -604,10 +611,13 @@ def vmap(fn: Callable[..., Any],
         tree_map_rngs(split_fn, rng_group) if split else rng_group
         for rng_group, split in zip(rng_groups, rng_splits))
 
-    @functools.partial(jax.vmap,
-                       in_axes=(variable_in_axes, rng_axes, in_axes),
-                       out_axes=(out_axes, variable_out_axes),
-                       axis_name=axis_name)
+    @functools.partial(
+        jax.vmap,
+        in_axes=(variable_in_axes, rng_axes, in_axes),
+        out_axes=(out_axes, variable_out_axes),
+        axis_name=axis_name,
+        axis_size=axis_size,
+        spmd_axis_name=spmd_axis_name)
     @functools.wraps(fn)
     def mapped(variable_groups, rng_groups, args):
       scope = scope_fn(variable_groups, rng_groups)
@@ -633,6 +643,7 @@ def scan(fn: Callable[..., Any],
          in_axes=0, out_axes=0,
          length: Optional[int] = None,
          reverse: bool = False,
+         unroll: int = 1,
          data_transform: Optional[Callable[..., Any]] = None,
          ) -> Callable[..., Any]:
   """A lifted version of ``jax.lax.scan``.
@@ -693,6 +704,8 @@ def scan(fn: Callable[..., Any],
     length: Specifies the number of loop iterations. This only needs
       to be specified if it cannot be derivied from the scan arguments.
     reverse: If true, scan from end to start in reverse order.
+    unroll: how many scan iterations to unroll within a single
+      iteration of a loop (default: 1).
     data_transform: optional function to transform raw variable and rng groups,
       intended for inline SPMD annotations.
 
@@ -715,13 +728,13 @@ def scan(fn: Callable[..., Any],
             init, *args):
     def find_length(axis, x):
       if axis is not axes_scan.broadcast:
-        leaves = jax.tree_leaves(x)
+        leaves = jax.tree_util.tree_leaves(x)
         if leaves:
           return leaves[0].shape[axis]
       return ()
     # split rngs
-    lengths = jax.tree_map(find_length, in_axes, args)
-    lengths = set(jax.tree_leaves(lengths))
+    lengths = jax.tree_util.tree_map(find_length, in_axes, args)
+    lengths = set(jax.tree_util.tree_leaves(lengths))
     if length is None and len(lengths) == 1:
       d_length, = lengths
     elif len(lengths) > 1:
@@ -739,7 +752,8 @@ def scan(fn: Callable[..., Any],
     @functools.partial(axes_scan.scan,
                        in_axes=(variable_in_axes, rng_axes, in_axes),
                        out_axes=(out_axes, variable_out_axes),
-                       length=length, reverse=reverse)
+                       length=length, reverse=reverse,
+                       unroll=unroll)
     def scanned(broadcast_vars, carry, scan_variable_groups, rng_groups, args):
       carry_vars, c = carry
       variable_groups = (broadcast_vars, carry_vars) + scan_variable_groups
@@ -867,10 +881,10 @@ def while_loop(cond_fn: Callable[[Scope, C], bool],
       name='while_loop')(scope)
 
 
-def cond(pred: Any, 
+def cond(pred: Any,
          true_fun: Callable[..., C], false_fun: Callable[..., C],
          scope: Scope, *operands,
-         variables: CollectionFilter = True, 
+         variables: CollectionFilter = True,
          rngs: PRNGSequenceFilter = True) -> C:
   """Lifted version of ``jax.lax.cond``.
 
@@ -903,7 +917,7 @@ def cond(pred: Any,
       The signature is (Scope, *operands) -> T.
     false_fun: The function evalauted when ``pred`` is `False`.
       The signature is (Scope, *operands) -> T.
-    scope: A Scope or Pytree of scopes to pass 
+    scope: A Scope or Pytree of scopes to pass
     *operands: The arguments passed to ``true_fun`` and ``false_fun``
     variables: The variable collections passed to the conditional
       branches (default: all)
@@ -932,6 +946,94 @@ def cond(pred: Any,
       name='cond')(scope)
 
 
+def switch(index: Any,
+           branches: Sequence[Callable[..., C]],
+           scope: Scope, *operands,
+           variables: CollectionFilter = True,
+           rngs: PRNGSequenceFilter = True) -> C:
+  """Lifted version of ``jax.lax.switch``.
+
+  The returned values from ``branches``
+  must have the same Pytree structure, shapes, and dtypes.
+  The variables created or updated inside the
+  branches must also have the same structure.
+  Note that this constraint is violated when
+  creating variables or submodules in only one branch.
+  Because initializing variables in just one branch
+  causes the paramater structure to be different.
+
+  Example::
+
+    def switch_example(scope, x, index):
+      scope.variable('state', 'a_count', lambda: 0)
+      scope.variable('state', 'b_count', lambda: 0)
+      scope.variable('state', 'c_count', lambda: 0)
+      def a_fn(scope, x):
+        scope.variable('state', 'a_count').value += 1
+        return scope.child(nn.dense)(x, 2)
+      def b_fn(scope, x):
+        scope.variable('state', 'b_count').value += 1
+        return -scope.child(nn.dense)(x, 2)
+      def c_fn(scope, x):
+        scope.variable('state', 'c_count').value += 1
+        return scope.child(nn.dense)(x, 2)
+      return lift.switch(index, [a_fn, b_fn, c_fn], scope, x)
+
+  If you want to have a different parameter structure for each branch
+  you should run all branche on initialization before calling switch::
+
+    def multihead_switch_example(scope, x, index):
+      def a_fn(scope, x):
+        x = scope.child(nn.dense)(x, 10)
+        x = scope.child(nn.dense)(x, 7)
+        return scope.child(nn.dense)(x, 5)
+      def b_fn(scope, x):
+        x = scope.child(nn.dense)(x, 11)
+        return scope.child(nn.dense)(x, 5)
+      def c_fn(scope, x):
+        return scope.child(nn.dense)(x, 5)
+
+      branches = [a_fn, b_fn, c_fn]
+
+      # run all branches on init
+      if scope.is_mutable_collection('params'):
+        for branch in branches:
+          _ = branch(scope, x)
+
+      return lift.switch(index, branches, scope, x)
+
+  Args:
+    index: Integer scalar type, indicating which branch function to apply.
+    branches: Sequence of functions to be applied based on index.
+      The signature of each function is (Scope, *operands) -> T.
+    scope: A Scope or Pytree of scopes to pass
+    *operands: The arguments passed to ``true_fun`` and ``false_fun``
+    variables: The variable collections passed to the conditional
+      branches (default: all)
+    rngs: The PRNG sequences passed to the conditionals (default: all)
+  Returns:
+    The result of the evaluated branch.
+  """
+
+  def inner(scope_fn, repack_fn,
+            variable_groups, rng_groups):
+    def branch_wrapper(branch_fn, *operands):
+      scope = scope_fn(variable_groups, rng_groups)
+      y = branch_fn(scope, *operands)
+      return y, repack_fn(scope)
+    pure_branches = [
+        functools.partial(branch_wrapper, branch_fn)
+        for branch_fn in branches]
+    return jax.lax.switch(index, pure_branches, *operands)
+
+  return pack(
+      inner,
+      (variables,),
+      (variables,),
+      (rngs,),
+      name='switch')(scope)
+
+
 def custom_vjp(fn: Callable[..., Any],
                forward_fn: Callable[..., Any],
                backward_fn: Callable[..., Any],
@@ -964,7 +1066,7 @@ def custom_vjp(fn: Callable[..., Any],
 
     def bwd(features, vjp_fn, y_t):
       input_t, params_t = vjp_fn(y_t)
-      params_t = jax.tree_map(jnp.sign, params_t)
+      params_t = jax.tree_util.tree_map(jnp.sign, params_t)
       return input_t, params_t
 
     dense_sign_grad = lift.custom_vjp(
@@ -1000,7 +1102,7 @@ def custom_vjp(fn: Callable[..., Any],
     def f_fwd(grad_variables, *args):
       nonlocal scopes_treedef
       scopes = scope_fn((grad_variables, other_variables), rng_groups)
-      scopes_treedef = jax.tree_structure(scopes)
+      scopes_treedef = jax.tree_util.tree_structure(scopes)
       y, res = forward_fn(scopes, *args)
       vars_out = repack_fn(scopes)
       return (y, vars_out), res
@@ -1176,7 +1278,7 @@ def jit(fn: Callable[..., Any],
     try:
       scope_fn = scope_fun
       repack_fn = repack_fun
-      scopes = jax.tree_leaves(scope_fn(variable_groups, rng_groups))
+      scopes = jax.tree_util.tree_leaves(scope_fn(variable_groups, rng_groups))
       mutable = tuple(_hashable_filter(scope.mutable) for scope in scopes)
       return jitted(mutable, variable_groups, rng_groups, *args)
     finally:
@@ -1251,22 +1353,6 @@ def remat_scan(
   return fn
 
 
-def named_call(fn: Callable[..., Any], name: str) -> Callable[..., Any]:
-  """Adds a name scope to `fn` during profiling."""
-  def inner(scope_fn, repack_fn, variable_groups, rng_groups, args, kwargs):
-    @functools.wraps(fn)
-    def named(variable_groups, rng_groups):
-      scope = scope_fn(variable_groups, rng_groups)
-      y = fn(scope, *args, **kwargs)
-      return y, repack_fn(scope)
-    named = jax.named_call(named, name=name)
-    return named(variable_groups, rng_groups)
-  lifted = pack(inner, (True,), (True,), (True,))
-  def wrapper(scope, *args, **kwargs):
-    return lifted(scope, args, kwargs)
-  return wrapper
-
-
 def _unzip2(xs):
   ys = tuple(zip(*xs))
   return ys if ys else ((), ())
