@@ -50,7 +50,7 @@ class Template(FlaxError):
 
 class FlaxError(Exception):
   def __init__(self, message):
-    error_page = 'https://flax.readthedocs.io/en/latest/flax.errors.html'
+    error_page = 'https://flax.readthedocs.io/en/latest/api_reference/flax.errors.html'
     module_name = self.__class__.__module__
     class_name = self.__class__.__name__
     error_msg = f'{message} ({error_page}#{module_name}.{class_name})'
@@ -551,24 +551,66 @@ class InvalidCheckpointError(FlaxError):
     super().__init__(f'Trying to save an outdated checkpoint at step: "{step}" and path: "{path}".')
 
 
+class GDACheckpointingRequiredError(FlaxError):
+  """To optimally save and restore a GlobalDeviceArray, use GlobalAsyncCheckpointManager.
+
+  You can create an GlobalAsyncCheckpointManager at top-level and pass it as
+  argument::
+
+    from jax.experimental.gda_serialization import serialization as gdas
+    gda_manager = gdas.GlobalAsyncCheckpointManager()
+    save_checkpoint(..., gda_manager=gda_manager)
+
+  """
+  def __init__(self, path, step):
+    super().__init__(
+        f'Checkpoint failed at step: "{step}" and path: "{path}": Target '
+        'contains a GlobalDeviceArray should be saved/restored with a '
+        'GlobalAsyncCheckpointManager.'
+    )
+
+
+class GDARestoreTargetRequiredError(FlaxError):
+  """Provide a valid target when restoring a checkpoint with GlobalDeviceArray.
+
+  To restore a checkpoint that contains a GlobalDeviceArray, make sure
+  to pass a valid ``target`` argument that contains GDAs with correct global
+  meshes and partition specs.
+
+  """
+
+  def __init__(self, path, step):
+    super().__init__(
+        f'Restore checkpoint failed at step: "{step}" and path: "{path}": '
+        'Checkpoints containing a GlobalDeviceArray need to be restored with '
+        'a valid target.'
+    )
+
+class GDARestoreDataCorruptedError(FlaxError):
+  """A GDA stored in Google Cloud Storage doesn't contain
+  a "commit_success.txt" file, which should be written at the end of the save.
+
+  Failure of finding it could indicate a corruption of your saved GDA data.
+  """
+
+  def __init__(self, step, path):
+    super().__init__(
+        f'Restore checkpoint failed at step: "{step}" on GlobalDeviceArray at '
+        f' "{path}": No "commit_success.txt" found on this GDA directory. '
+        'Was its save halted before completion?'
+    )
+
+
 #################################################
 # transforms.py errors                          #
 #################################################
 
 class TransformedMethodReturnValueError(FlaxError):
-  """
-  Transformed Module methods cannot return other Modules or Variables.
+  """Transformed Module methods cannot return other Modules or Variables."""
 
-  This commonly occurs when ``@nn.named_call`` is automatically applied to
-  helper constructor methods when profiling is enabled (``FLAX_PROFILE=true``
-  environment variable or via ``nn.enable_named_call()``), and can be mitigated
-  by using the ``@nn.nowrap`` decorator to prevent automatic wrapping.
-  """
   def __init__(self, name):
     super().__init__(
-      f'Transformed module method {name} cannot return Modules or Variables. '
-      f'For helper constructor methods use the @nn.nowrap decorator to prevent '
-      f'decoration by the automatic named_call transform.')
+        f'Transformed module method {name} cannot return Modules or Variables.')
 
 
 class TransformTargetError(FlaxError):
