@@ -17,7 +17,7 @@
 
 # pylint: disable=unused-import
 # re-export activation functions from jax.nn
-from typing import Any
+from typing import Any, Optional
 
 from flax.linen.module import compact
 from flax.linen.module import Module
@@ -27,12 +27,15 @@ from jax.nn import elu
 from jax.nn import gelu
 from jax.nn import glu
 from jax.nn import hard_sigmoid
+from jax.nn import hard_silu
 from jax.nn import hard_swish
 from jax.nn import hard_tanh
 from jax.nn import leaky_relu
 from jax.nn import log_sigmoid
 from jax.nn import log_softmax
+from jax.nn import logsumexp
 from jax.nn import normalize
+from jax.nn import one_hot
 from jax.nn import relu
 from jax.nn import relu6
 from jax.nn import selu
@@ -49,14 +52,18 @@ from jax.numpy import tanh
 
 
 Array = Any
+Dtype = Any
 
 
 class PReLU(Module):
   """Parametric Rectified Linear Unit (PReLU) activation function.
 
   Attributes:
-    negative_slope_init: the value to initialize the negative slope.
+    param_dtype: the dtype passed to parameter initializers (default: float32).
+    negative_slope_init: the value to initialize the negative slope
+      (default 0.01).
   """
+  param_dtype: Dtype = jnp.float32
   negative_slope_init: float = 0.01
 
   @compact
@@ -71,6 +78,6 @@ class PReLU(Module):
     """
     negative_slope = self.param(
         'negative_slope',
-        lambda k: jnp.asarray(self.negative_slope_init, jnp.float32))
+        lambda k: jnp.asarray(self.negative_slope_init, self.param_dtype))
     return jnp.where(inputs >= 0, inputs,
                      jnp.asarray(negative_slope, inputs.dtype) * inputs)
