@@ -44,7 +44,7 @@ default_kernel_init = lecun_normal()
 
 def _normalize_axes(axes: Tuple[int, ...], ndim: int) -> Tuple[int, ...]:
   # A tuple by convention. len(axes_tuple) then also gives the rank efficiently.
-  return tuple(sorted([ax if ax >= 0 else ndim + ax for ax in axes]))
+  return tuple(sorted(ax if ax >= 0 else ndim + ax for ax in axes))
 
 
 def _canonicalize_tuple(x: Union[Sequence[int], int]) -> Tuple[int, ...]:
@@ -113,12 +113,12 @@ class DenseGeneral(Module):
                                 for _ in range(size_batch_dims)], axis=0)
       return jnp.reshape(kernel, shape)
 
-    batch_shape = tuple([inputs.shape[ax] for ax in batch_dims])
+    batch_shape = tuple(inputs.shape[ax] for ax in batch_dims)
     # batch and non-contracting dims of input with 1s for batch dims.
     expanded_batch_shape = tuple(
         inputs.shape[ax] if ax in batch_dims else 1
         for ax in range(inputs.ndim) if ax not in axis)
-    kernel_shape = tuple([inputs.shape[ax] for ax in axis]) + features
+    kernel_shape = tuple(inputs.shape[ax] for ax in axis) + features
     kernel = self.param('kernel', kernel_init_wrap, batch_shape + kernel_shape,
                         self.param_dtype)
 
@@ -252,7 +252,8 @@ class _Conv(Module):
       high)` integer pairs that give the padding to apply before and after each
       spatial dimension. A single int is interpeted as applying the same padding
       in all dims and passign a single int in a sequence causes the same padding
-      to be used on both sides.
+      to be used on both sides. `'CAUSAL'` padding for a 1D convolution will
+      left-pad the convolution axis, resulting in same-sized output.
     input_dilation: an integer or a sequence of `n` integers, giving the
       dilation factor to apply in each spatial dimension of `inputs`
       (default: 1). Convolution with input dilation `d` is equivalent to
@@ -264,6 +265,8 @@ class _Conv(Module):
     feature_group_count: integer, default 1. If specified divides the input
       features into groups.
     use_bias: whether to add a bias to the output (default: True).
+    mask: Optional mask for the weights during masked convolution. The mask must
+          be the same shape as the convolution weight matrix.
     dtype: the dtype of the computation (default: infer from input and params).
     param_dtype: the dtype passed to parameter initializers (default: float32).
     precision: numerical precision of the computation see `jax.lax.Precision`
@@ -279,6 +282,7 @@ class _Conv(Module):
   kernel_dilation: Union[None, int, Sequence[int]] = 1
   feature_group_count: int = 1
   use_bias: bool = True
+  mask: Optional[Array] = None
   dtype: Optional[Dtype] = None
   param_dtype: Dtype = jnp.float32
   precision: PrecisionLike = None
@@ -302,11 +306,17 @@ class _Conv(Module):
     """Applies a (potentially unshared) convolution to the inputs.
 
     Args:
-      inputs: input data with dimensions (batch, spatial_dims..., features).
-        This is the channels-last convention, i.e. NHWC for a 2d convolution
-        and NDHWC for a 3D convolution. Note: this is different from the input
-        convention used by `lax.conv_general_dilated`, which puts the spatial
-        dimensions last.
+      inputs: input data with dimensions (*batch_dims, spatial_dims...,
+        features). This is the channels-last convention, i.e. NHWC for a 2d
+        convolution and NDHWC for a 3D convolution. Note: this is different from
+        the input convention used by `lax.conv_general_dilated`, which puts the
+        spatial dimensions last.
+        Note: If the input has more than 1 batch dimension, all batch dimensions
+        are flattened into a single dimension for the convolution and restored
+        before returning.  In some cases directly vmap'ing the layer may yield
+        better performance than this default flattening approach.  If the input
+        lacks a batch dimension it will be added for the convolution and removed
+        n return, an allowance made to enable writing single-example code.
 
     Returns:
       The convolved data.
@@ -329,10 +339,14 @@ class _Conv(Module):
         return (x,) * len(kernel_size)
       return tuple(x)
 
-    is_single_input = False
-    if inputs.ndim == len(kernel_size) + 1:
-      is_single_input = True
-      inputs = jnp.expand_dims(inputs, axis=0)
+    # Combine all input batch dimensions into a single leading batch axis.
+    num_batch_dimensions = inputs.ndim - (len(kernel_size) + 1)
+    if num_batch_dimensions != 1:
+      input_batch_shape = inputs.shape[:num_batch_dimensions]
+      total_batch_size = int(np.prod(input_batch_shape))
+      flat_input_shape = (
+          (total_batch_size,) + inputs.shape[num_batch_dimensions:])
+      inputs = jnp.reshape(inputs, flat_input_shape)
 
     # self.strides or (1,) * (inputs.ndim - 2)
     strides = maybe_broadcast(self.strides)
@@ -349,6 +363,14 @@ class _Conv(Module):
               [(0, 0)])
       inputs = jnp.pad(inputs, pads, mode='wrap')
       padding_lax = 'VALID'
+    elif padding_lax == 'CAUSAL':
+      if len(kernel_size) != 1:
+        raise ValueError(
+            'Causal padding is only implemented for 1D convolutions.')
+      left_pad = kernel_dilation[0] * (kernel_size[0] - 1)
+      pads = [(0, 0), (left_pad, 0), (0, 0)]
+      inputs = jnp.pad(inputs, pads)
+      padding_lax = 'VALID'
 
     dimension_numbers = _conv_dimension_numbers(inputs.shape)
     in_features = jnp.shape(inputs)[-1]
@@ -386,9 +408,16 @@ class _Conv(Module):
       kernel_shape = conv_output_shape[1:-1] + (np.prod(kernel_size) *
                                                 in_features, self.features)
 
+    if self.mask is not None and self.mask.shape != kernel_shape:
+      raise ValueError('Mask needs to have the same shape as weights. '
+                       f'Shapes are: {self.mask.shape}, {kernel_shape}')
+
     kernel = self.param('kernel', self.kernel_init, kernel_shape,
                         self.param_dtype)
 
+    if self.mask is not None:
+      kernel *= self.mask
+
     if self.use_bias:
       if self.shared_weights:
         # One bias weight per output channel, shared between pixels.
@@ -431,8 +460,9 @@ class _Conv(Module):
       bias = bias.reshape((1,) * (y.ndim - bias.ndim) + bias.shape)
       y += bias
 
-    if is_single_input:
-      y = jnp.squeeze(y, axis=0)
+    if num_batch_dimensions != 1:
+      output_shape = input_batch_shape + y.shape[1:]
+      y = jnp.reshape(y, output_shape)
     return y
 
 
@@ -472,6 +502,8 @@ class ConvTranspose(Module):
       kernel. Convolution with kernel dilation is also known as 'atrous
       convolution'.
     use_bias: whether to add a bias to the output (default: True).
+    mask: Optional mask for the weights during masked convolution. The mask must
+          be the same shape as the convolution weight matrix.
     dtype: the dtype of the computation (default: infer from input and params).
     param_dtype: the dtype passed to parameter initializers (default: float32).
     precision: numerical precision of the computation see `jax.lax.Precision`
@@ -485,6 +517,7 @@ class ConvTranspose(Module):
   padding: PaddingLike = 'SAME'
   kernel_dilation: Optional[Sequence[int]] = None
   use_bias: bool = True
+  mask: Optional[Array] = None
   dtype: Dtype = None
   param_dtype: Dtype = jnp.float32
   precision: PrecisionLike = None
@@ -498,11 +531,17 @@ class ConvTranspose(Module):
     Behaviour mirrors of `jax.lax.conv_transpose`.
 
     Args:
-      inputs: input data with dimensions (batch, spatial_dims..., features).
-        This is the channels-last convention, i.e. NHWC for a 2d convolution
-        and NDHWC for a 3D convolution. Note: this is different from the input
-        convention used by `lax.conv_general_dilated`, which puts the spatial
-        dimensions last.
+      inputs: input data with dimensions (*batch_dims, spatial_dims...,
+        features). This is the channels-last convention, i.e. NHWC for a 2d
+        convolution and NDHWC for a 3D convolution. Note: this is different from
+        the input convention used by `lax.conv_general_dilated`, which puts the
+        spatial dimensions last.
+        Note: If the input has more than 1 batch dimension, all batch dimensions
+        are flattened into a single dimension for the convolution and restored
+        before returning.  In some cases directly vmap'ing the layer may yield
+        better performance than this default flattening approach.  If the input
+        lacks a batch dimension it will be added for the convolution and removed
+        n return, an allowance made to enable writing single-example code.
 
     Returns:
       The convolved data.
@@ -513,19 +552,31 @@ class ConvTranspose(Module):
     else:
       kernel_size = self.kernel_size
 
-    is_single_input = False
-    if inputs.ndim == len(kernel_size) + 1:
-      is_single_input = True
-      inputs = jnp.expand_dims(inputs, axis=0)
+    # Combine all input batch dimensions into a single leading batch axis.
+    num_batch_dimensions = inputs.ndim - (len(kernel_size) + 1)
+    if num_batch_dimensions != 1:
+      input_batch_shape = inputs.shape[:num_batch_dimensions]
+      total_batch_size = int(np.prod(input_batch_shape))
+      flat_input_shape = (
+          (total_batch_size,) + inputs.shape[num_batch_dimensions:])
+      inputs = jnp.reshape(inputs, flat_input_shape)
 
     strides: Tuple[int, ...]
     strides = self.strides or (1,) * (inputs.ndim - 2)
 
     in_features = jnp.shape(inputs)[-1]
     kernel_shape = kernel_size + (in_features, self.features)
+
+    if self.mask is not None and self.mask.shape != kernel_shape:
+      raise ValueError('Mask needs to have the same shape as weights. '
+                       f'Shapes are: {self.mask.shape}, {kernel_shape}')
+
     kernel = self.param('kernel', self.kernel_init, kernel_shape,
                         self.param_dtype)
 
+    if self.mask is not None:
+      kernel *= self.mask
+
     padding_lax = canonicalize_padding(self.padding, len(kernel_size))
     if padding_lax == 'CIRCULAR':
       padding_lax = 'VALID'
@@ -573,7 +624,7 @@ class ConvTranspose(Module):
       total_pad = [
           ((size_diff + 1) // 2, size_diff // 2) for size_diff in size_diffs
       ]
-      y = np.pad(y, [(0, 0)] + total_pad + [(0, 0)])
+      y = jnp.pad(y, [(0, 0)] + total_pad + [(0, 0)])
       # Wrap the result periodically around each spatial dimension,
       # one by one.
       for i in range(1, y.ndim - 1):
@@ -581,10 +632,13 @@ class ConvTranspose(Module):
                       y.shape[i + 1:])
         y = y.sum(axis=i)
 
-    if is_single_input:
-      y = jnp.squeeze(y, axis=0)
     if self.use_bias:
       y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
+
+    if num_batch_dimensions != 1:
+      output_shape = input_batch_shape + y.shape[1:]
+      y = jnp.reshape(y, output_shape)
+
     return y
 
 
