@@ -20,22 +20,25 @@ import functools
 import inspect
 import threading
 import typing
-from typing import (Any, Callable, Dict, Iterable, List, Optional,
-                    Set, Tuple, Type, TypeVar, Union, overload)
-from typing_extensions import dataclass_transform  # pytype: disable=not-supported-yet
 import weakref
+from typing import (Any, Callable, Dict, Generic, Iterable, List, Optional,
+                    Sequence, Set, Tuple, Type, TypeVar, Union, overload)
 
-from flax import config
-from flax import core
-from flax import errors
-from flax import serialization
-from flax import traceback_util
-from flax import traverse_util
+import jax
+import numpy as np
+from typing_extensions import \
+    dataclass_transform  # pytype: disable=not-supported-yet
+
+from flax import (config, core, errors, serialization, traceback_util,
+                  traverse_util)
 from flax.core import Scope
 from flax.core.frozen_dict import FrozenDict
-from flax.core.scope import (CollectionFilter, DenyList, FrozenVariableDict,  # pylint: disable=g-multiple-import
-                             Variable, VariableDict, union_filters)
-import jax
+from flax.core.scope import (  # pylint: disable=g-multiple-import
+    CollectionFilter, DenyList, FrozenVariableDict, Variable, VariableDict,
+    union_filters)
+from flax.ids import uuid
+from flax.linen import summary
+
 
 
 traceback_util.register_exclusion(__file__)
@@ -139,15 +142,26 @@ _unspecified_parent = _Sentinel()
 _use_named_call = config.flax_profile
 
 
+def _derive_profiling_name(module, fn):
+  def _get_fn_name(fn):
+    if isinstance(fn, functools.partial):
+      return _get_fn_name(fn.func)
+    return fn.__name__
+  fn_name = _get_fn_name(fn)
+  method_suffix = f'.{fn_name}' if fn_name != '__call__' else ''
+  module_name = module.name or module.__class__.__name__
+  return f'{module_name}{method_suffix}'
+
+
 def enable_named_call():
   """Enables named call wrapping for labelling profile traces.
 
   When named call wrapping is enabled all JAX ops executed in a Module
-  will be wrapped with ``jax.named_call``. The ``Module`` class name will
+  will be run under ``jax.named_scope``. The ``Module`` class name will
   show up around the operations belonging to that Module in the
   Tensorboard profiling UI, simplifying the profiling process.
 
-  Note that ``jax.named_call`` only works for
+  Note that ``jax.named_scope`` only works for
   compiled functions (e.g.: using jax.jit or jax.pmap).
   """
   global _use_named_call
@@ -278,9 +292,6 @@ def nowrap(fun: _CallableT) -> _CallableT:
   with the state handler or a separate named_call transform.
 
   This is needed in several concrete instances:
-   - if you have a helper method that returns Modules or Variables to prevent
-     it from being functionalized by named_call. (Functionalized methods
-     can't return Modules/Variables.)
    - if you're subclassing a method like Module.param and don't want this
      overriden core function decorated with the state management wrapper.
    - If you want a method to be callable from an unbound Module (e.g.: a
@@ -382,7 +393,8 @@ def _get_unbound_fn(method_or_fn: Callable[..., Any]) -> Callable[..., Any]:
   Returns:
     An unbound version of input function.
   """
-  if inspect.ismethod(method_or_fn):
+  if (inspect.ismethod(method_or_fn) and
+      isinstance(method_or_fn.__self__, Module)):  # pytype: disable=attribute-error
     method_or_fn = method_or_fn.__func__  # pytype: disable=attribute-error
 
   # The method should be callable, and it should have at least one argument
@@ -490,7 +502,7 @@ class Module:
     from flax import linen as nn
 
     class Module(nn.Module):
-      features: Tuple[int] = (16, 4)
+      features: Tuple[int, ...] = (16, 4)
 
       def setup(self):
         self.dense1 = Dense(self.features[0])
@@ -530,7 +542,7 @@ class Module:
     cls._wrap_module_methods()
     # Set empty class defaults.
     cls._state = _uninitialized_module_internal_state
-    cls.scope = None
+    cls.scope: Optional[Scope] = None
 
   @classmethod
   def _customized_dataclass_transform(cls):
@@ -604,12 +616,7 @@ class Module:
       method = getattr(cls, key)
       if hasattr(method, 'nowrap'):
         continue
-      wrapped_method = wrap_method_once(method)
-      if key != 'setup':
-        # We import named_call at runtime to avoid a circular import issue.
-        from flax.linen.transforms import named_call  # pylint: disable=g-import-not-at-top
-        wrapped_method = named_call(wrapped_method, force=False)
-      setattr(cls, key, wrapped_method)
+      setattr(cls, key, wrap_method_once(method))
     return cls
 
   def _call_wrapped_method(self, fun, args, kwargs):
@@ -645,7 +652,11 @@ class Module:
       self._state.in_compact_method = True
     _context.module_stack.append(self)
     try:
-      y = fun(self, *args, **kwargs)
+      if _use_named_call:
+        with jax.named_scope(_derive_profiling_name(self, fun)):
+          y = fun(self, *args, **kwargs)
+      else:
+        y = fun(self, *args, **kwargs)
       if _context.capture_stack:
         filter_fn = _context.capture_stack[-1]
         if filter_fn and filter_fn(self, fun_name):
@@ -714,7 +725,7 @@ class Module:
       raise AttributeError(
           f'"{self.__class__.__name__}" object has no attribute "{name}"')
 
-  def __dir__(self) -> Iterable[str]:
+  def __dir__(self) -> List[str]:
     """Call setup() before listing attributes."""
     self._try_setup()
     return object.__dir__(self)  # pytype: disable=attribute-error
@@ -727,6 +738,7 @@ class Module:
     # initialization, attach this Module as a submodule of a parent, or bind
     # this Module at the top-level to variables and rngs.
 
+    object.__setattr__(self, '_id', uuid())
     object.__setattr__(self, '_state', _ModuleInternalState())
 
     # Typically we set the parent based on the dynamic module context.
@@ -824,7 +836,9 @@ class Module:
           # Module was passed from outside. It needs to be cloned.
           # Outside modules are named by attachment, not an outer name.
           object.__setattr__(subvalue, 'name', None)
-          key = id(subvalue)
+          # Preserve sharing-by-reference relationships during adoption
+          # via cache keyed on unique instance ids.
+          key = subvalue._id
           if key not in cache:
             cache[key] = subvalue.clone()
           subvalue = cache[key]
@@ -873,7 +887,7 @@ class Module:
   def _name_taken(self,
                   name: str,
                   module: 'Module' = None,
-                  reuse_scopes : bool = False) -> bool:
+                  reuse_scopes: bool = False) -> bool:
     if name in _all_names_on_object(self):
       val = getattr(self, name, None)
       if module is not None and val is module:
@@ -1030,6 +1044,20 @@ class Module:
       raise ValueError("Can't use RNGs on unbound modules")
     return self.scope.make_rng(name)
 
+  def is_initializing(self) -> bool:
+    """Returns True if running under self.init(...) or nn.init(...)().
+
+    This is a helper method to handle the common case of simple initialization
+    where we wish to have setup logic occur when only called under
+    ``module.init`` or ``nn.init``.  For more complicated multi-phase
+    initialization scenarios it is better to test for the mutability of
+    particular variable collections or for the presence of particular
+    variables that potentially need to be initialized.
+    """
+    if self.scope is None:
+      raise ValueError("Can't check if running under init() on unbound modules")
+    return self.scope.get_flag('initializing', False)
+
   @traceback_util.api_boundary
   def bind(self,
            variables: VariableDict,
@@ -1039,14 +1067,14 @@ class Module:
     """Creates an interactive Module instance by binding variables and RNGs.
 
     ``bind`` provides an "interactive" instance of a Module directly without
-    transforming a function with ``apply``. This is particalary useful for debugging
+    transforming a function with ``apply``. This is particularly useful for debugging
     and interactive use cases like notebooks where a function would limit the ability
-    split up code into different cells.
+    to split up code into different cells.
 
     Once the variables (and optionally RNGs) are bound to a ``Module`` it becomes a
     stateful object. Note that idiomatic JAX is functional and therefore an interactive
-    instance does not mix well well with vanilla JAX APIs. ``bind()`` should only be used
-    for interactive experimentation, and in all other cases we strongly encourage
+    instance does not mix well with vanilla JAX APIs. ``bind()`` should only be used
+    for interactive experimentation, and in all other cases we strongly encourage users
     to use ``apply()`` instead.
 
     Example::
@@ -1096,8 +1124,7 @@ class Module:
             rngs: Optional[RNGSequences] = None,
             method: Optional[Callable[..., Any]] = None,
             mutable: CollectionFilter = False,
-            capture_intermediates: Union[bool, Callable[['Module', str],
-                                                        bool]] = False,
+            capture_intermediates: Union[bool, Callable[['Module', str], bool]] = False,
             **kwargs) -> Union[Any, Tuple[Any, FrozenVariableDict]]:
     """Applies a module method to variables and returns output and modified variables.
 
@@ -1155,7 +1182,8 @@ class Module:
     method = _get_unbound_fn(method)
     return apply(
         method, self,
-        mutable=mutable, capture_intermediates=capture_intermediates
+        mutable=mutable,
+        capture_intermediates=capture_intermediates,
     )(variables, *args, **kwargs, rngs=rngs)
 
   @traceback_util.api_boundary
@@ -1164,6 +1192,7 @@ class Module:
                        *args,
                        method: Optional[Callable[..., Any]] = None,
                        mutable: CollectionFilter = DenyList('intermediates'),
+                       capture_intermediates: Union[bool, Callable[['Module', str], bool]] = False,
                        **kwargs) -> Tuple[Any, FrozenVariableDict]:
     """Initializes a module method with variables and returns output and modified variables.
 
@@ -1177,6 +1206,12 @@ class Module:
         ``str``: The name of a single mutable collection. ``list``: A
         list of names of mutable collections. By default all collections
         except "intermediates" are mutable.
+      capture_intermediates: If `True`, captures intermediate return values
+        of all Modules inside the "intermediates" collection. By default only
+        the return values of all ``__call__`` methods are stored. A function can
+        be passed to change the filter behavior. The filter function takes
+        the Module instance and method name and returns a bool indicating
+        whether the output of that method invocation should be stored.
       **kwargs: Keyword arguments passed to the init function.
     Returns:
       `(output, vars)``, where ``vars`` are is a dict of the modified
@@ -1188,8 +1223,15 @@ class Module:
             'RNGs should be of shape (2,) or KeyArray in Module '
             f'{self.__class__.__name__}, but rngs are: {rngs}')
       rngs = {'params': rngs}
-    return self.apply(
-        {}, *args, rngs=rngs, method=method, mutable=mutable, **kwargs)
+    if method is None:
+      method = self.__call__
+    method = _get_unbound_fn(method)
+    return init_with_output(
+        method,
+        self,
+        mutable=mutable,
+        capture_intermediates=capture_intermediates
+    )(rngs, *args, **kwargs)
 
   @traceback_util.api_boundary
   def init(self,
@@ -1197,6 +1239,7 @@ class Module:
            *args,
            method: Optional[Callable[..., Any]] = None,
            mutable: CollectionFilter = DenyList('intermediates'),
+           capture_intermediates: Union[bool, Callable[['Module', str], bool]] = False,
            **kwargs) -> FrozenVariableDict:
     """Initializes a module method with variables and returns modified variables.
 
@@ -1217,13 +1260,23 @@ class Module:
         ``str``: The name of a single mutable collection. ``list``: A
         list of names of mutable collections. By default all collections
         except "intermediates" are mutable.
+      capture_intermediates: If `True`, captures intermediate return values
+        of all Modules inside the "intermediates" collection. By default only
+        the return values of all ``__call__`` methods are stored. A function can
+        be passed to change the filter behavior. The filter function takes
+        the Module instance and method name and returns a bool indicating
+        whether the output of that method invocation should be stored.
       **kwargs: Keyword arguments passed to the init function.
     Returns:
       The initialized variable dict.
     """
     _, v_out = self.init_with_output(
-        rngs, *args,
-        method=method, mutable=mutable, **kwargs)
+        rngs,
+        *args,
+        method=method,
+        mutable=mutable,
+        capture_intermediates=capture_intermediates,
+        **kwargs)
     return v_out
 
   @property
@@ -1353,6 +1406,98 @@ class Module:
     self.scope.put_variable(col, name, xs)
     return True
 
+  def tabulate(
+    self,
+    rngs: Union[PRNGKey, RNGSequences],
+    *args,
+    method: Optional[Callable[..., Any]] = None,
+    mutable: CollectionFilter = True,
+    depth: Optional[int] = None,
+    exclude_methods: Sequence[str] = (),
+    **kwargs) -> str:
+    """Creates a summary of the Module represented as a table.
+
+    This method has the same signature as `init`, but instead of returning
+    the variables, it returns the string summarizing the Module in a table.
+    `tabulate` uses `jax.eval_shape` to run the forward computation without
+    consuming any FLOPs or allocating memory.
+
+    Example::
+
+      import jax
+      import jax.numpy as jnp
+      import flax.linen as nn
+
+      class Foo(nn.Module):
+          @nn.compact
+          def __call__(self, x):
+              h = nn.Dense(4)(x)
+              return nn.Dense(2)(h)
+
+      x = jnp.ones((16, 9))
+
+      print(Foo().tabulate(jax.random.PRNGKey(0), x))
+
+
+    This gives the following output::
+
+                         Foo Summary
+      ┏━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
+      ┃ path    ┃ outputs       ┃ params               ┃
+      ┡━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
+      │ Inputs  │ float32[16,9] │                      │
+      ├─────────┼───────────────┼──────────────────────┤
+      │ Dense_0 │ float32[16,4] │ bias: float32[4]     │
+      │         │               │ kernel: float32[9,4] │
+      │         │               │                      │
+      │         │               │ 40 (160 B)           │
+      ├─────────┼───────────────┼──────────────────────┤
+      │ Dense_1 │ float32[16,2] │ bias: float32[2]     │
+      │         │               │ kernel: float32[4,2] │
+      │         │               │                      │
+      │         │               │ 10 (40 B)            │
+      ├─────────┼───────────────┼──────────────────────┤
+      │ Foo     │ float32[16,2] │                      │
+      ├─────────┼───────────────┼──────────────────────┤
+      │         │         Total │ 50 (200 B)           │
+      └─────────┴───────────────┴──────────────────────┘
+
+                Total Parameters: 50 (200 B)
+
+    **Note**: rows order in the table does not represent execution order,
+    instead it aligns with the order of keys in `variables` which are sorted
+    alphabetically.
+
+    Args:
+      rngs: The rngs for the variable collections.
+      *args: The arguments to the forward computation.
+      method: An optional method. If provided, applies this method. If not
+        provided, applies the ``__call__`` method.
+      mutable: Can be bool, str, or list. Specifies which collections should be
+        treated as mutable: ``bool``: all/no collections are mutable.
+        ``str``: The name of a single mutable collection. ``list``: A
+        list of names of mutable collections. By default all collections
+        except 'intermediates' are mutable.
+      depth: controls how many submodule deep the summary can go. By default its
+        `None` which means no limit. If a submodule is not shown because of the
+        depth limit, its parameter count and bytes will be added to the row of
+        its first shown ancestor such that the sum of all rows always adds up to
+        the total number of parameters of the Module.
+      exclude_methods: A sequence of strings that specifies which methods should
+        be ignored. In case a module calls a helper method from its main method,
+        use this argument to exclude the helper method from the summary to avoid
+        ambiguity.
+      **kwargs: keyword arguments to pass to the forward computation.
+
+    Returns:
+      A string summarizing the Module.
+    """
+
+    tabulate_fn = summary.tabulate(self, rngs, method=method,
+                                   mutable=mutable, depth=depth,
+                                   exclude_methods=exclude_methods)
+    return tabulate_fn(*args, **kwargs)
+
 
 def merge_param(name: str, a: Optional[T], b: Optional[T]) -> T:
   """Merges construction and call time argument.
@@ -1392,7 +1537,8 @@ def merge_param(name: str, a: Optional[T], b: Optional[T]) -> T:
 @traceback_util.api_boundary
 def apply(fn: Callable[..., Any], module: Module,
           mutable: CollectionFilter = False,
-          capture_intermediates: Union[bool, Callable[[Module, str], bool]] = False) -> Callable[..., Any]:
+          capture_intermediates: Union[bool, Callable[[Module, str], bool]] = False,
+          ) -> Callable[..., Any]:
   """Creates an apply function to call ``fn`` with a bound module.
 
   Unlike ``Module.apply`` this function returns a new function with the signature
@@ -1451,6 +1597,7 @@ def apply(fn: Callable[..., Any], module: Module,
 @traceback_util.api_boundary
 def init_with_output(fn: Callable[..., Any], module: Module,
                      mutable: CollectionFilter = DenyList('intermediates'),
+                     capture_intermediates: Union[bool, Callable[[Module, str], bool]] = False,
                      ) -> Callable[..., Tuple[Any, FrozenVariableDict]]:
   """Creates an init function to call ``fn`` with a bound module that also returns the function outputs.
 
@@ -1484,18 +1631,34 @@ def init_with_output(fn: Callable[..., Any], module: Module,
       ``str``: The name of a single mutable collection. ``list``: A
       list of names of mutable collections. By default all collections
       except "intermediates" are mutable.
+    capture_intermediates: If `True`, captures intermediate return values
+      of all Modules inside the "intermediates" collection. By default only
+      the return values of all `__call__` methods are stored. A function can
+      be passed to change the filter behavior. The filter function takes
+      the Module instance and method name and returns a bool indicating
+      whether the output of that method invocation should be stored.
   Returns:
     The init function wrapping ``fn``.
   """
   @functools.wraps(fn)
   def scope_fn(scope, *args, **kwargs):
-    return fn(module.clone(parent=scope), *args, **kwargs)
+    _context.capture_stack.append(capture_intermediates)
+    try:
+      return fn(module.clone(parent=scope), *args, **kwargs)
+    finally:
+      _context.capture_stack.pop()
+
+  if capture_intermediates is True:  # pylint: disable=g-bool-id-comparison
+    capture_intermediates = capture_call_intermediates
+  if capture_intermediates:
+    mutable = union_filters(mutable, 'intermediates')
   return core.init(scope_fn, mutable=mutable)
 
 
 @traceback_util.api_boundary
 def init(fn: Callable[..., Any], module: Module,
          mutable: CollectionFilter = DenyList('intermediates'),
+         capture_intermediates: Union[bool, Callable[[Module, str], bool]] = False,
          ) -> Callable[..., FrozenVariableDict]:
   """Creates an init function to call ``fn`` with a bound module.
 
@@ -1529,10 +1692,16 @@ def init(fn: Callable[..., Any], module: Module,
       ``str``: The name of a single mutable collection. ``list``: A
       list of names of mutable collections. By default all collections
       except "intermediates" are mutable.
+    capture_intermediates: If `True`, captures intermediate return values
+      of all Modules inside the "intermediates" collection. By default only
+      the return values of all `__call__` methods are stored. A function can
+      be passed to change the filter behavior. The filter function takes
+      the Module instance and method name and returns a bool indicating
+      whether the output of that method invocation should be stored.
   Returns:
     The init function wrapping ``fn``.
   """
-  init_fn = init_with_output(fn, module, mutable)
+  init_fn = init_with_output(fn, module, mutable, capture_intermediates)
   @functools.wraps(init_fn)
   def init_wrapper(*args, **kwargs):
     return init_fn(*args, **kwargs)[1]
