@@ -30,6 +30,7 @@ logical axis metadata to the underlying Lifted transformations.
 
 import collections
 import contextlib
+import enum
 import functools
 import re
 import threading
@@ -48,15 +49,14 @@ import jax
 from jax.experimental import maps
 from jax.experimental import pjit
 
-
 # Real types and dummy aliases for documentation
 LogicalRules = Sequence[Tuple[str, Union[str, Tuple[str], None]]]
 Array = Any  # pylint: disable=invalid-name
 ArrayPytree = Any  # pylint: disable=invalid-name
+LogicalPartitionSpec = Any  # pylint: disable=invalid-name
 LogicalPartitionSpecPytree = Any  # pylint: disable=invalid-name
 PartitionSpecPytree = Any  # pylint: disable=invalid-name
 
-
 # Dynamic Axis Mapping Context
 # ------------------------------------------------------------------------------
 
@@ -112,21 +112,52 @@ class _UnassignedAxis:
   def __bool__(self):
     return False
 
+
 _unassigned_axis = _UnassignedAxis()
 
 
 def _mesh_assignment_free(new_assignment, existing_assignments):
   """Determines if a given mesh axis has already been assigned."""
-  new = set(jax.tree_leaves(new_assignment))
-  existing = set(jax.tree_leaves(existing_assignments))
+  new = set(jax.tree_util.tree_leaves(new_assignment))
+  existing = set(jax.tree_util.tree_leaves(existing_assignments))
   if existing.intersection(new):
     return False
   return True
 
 
-def logical_to_mesh_axes(array_dim_names: Optional[Sequence[Optional[str]]],
-                         rules: Optional[LogicalRules] = None,
-                         ) -> pjit.PartitionSpec:
+def _logical_to_mesh_axes(
+    array_dim_names: Optional[Sequence[Optional[str]]],
+    rules: Optional[LogicalRules] = None,
+) -> Optional[List[Union[_UnassignedAxis, None, str, Tuple[str]]]]:
+  """Same as logical_to_mesh_axes, but doesn't fill in _unassigned_axis."""
+  if array_dim_names is None:
+    return None
+  if rules is None:
+    rules = _axis_rules.rules
+  axis_name_counts = collections.Counter(array_dim_names)
+  dups = tuple(
+      k for k, v in axis_name_counts.items() if v > 1 and k is not None)
+  if dups:
+    raise ValueError(
+        f'Unsupported: Dimensions {dups} occur more than once in array names.')
+  if not isinstance(rules, (tuple, list)):
+    raise ValueError('Unknown axis rule specification type.')
+  # We assign mesh axes using a priority based ruleset over logical axis names.
+  result: List[Union[_UnassignedAxis, None, str, Tuple[str]]]
+  result = [_unassigned_axis] * len(array_dim_names)
+  for rule_model_name, rule_mesh_names in rules:
+    if rule_model_name in array_dim_names:
+      pos = array_dim_names.index(rule_model_name)
+      if (_mesh_assignment_free(rule_mesh_names, result) and
+          result[pos] == _unassigned_axis):
+        result[pos] = rule_mesh_names
+  return result
+
+
+def logical_to_mesh_axes(
+    array_dim_names: Optional[Sequence[Optional[str]]],
+    rules: Optional[LogicalRules] = None,
+) -> pjit.PartitionSpec:
   """Compute layout for an array.
 
   The rules are in order of precedence, and consist of pairs:
@@ -159,27 +190,9 @@ def logical_to_mesh_axes(array_dim_names: Optional[Sequence[Optional[str]]],
   Returns:
     PartitionSpec for the parameter.
   """
-  if array_dim_names is None:
+  result = _logical_to_mesh_axes(array_dim_names, rules)
+  if result is None:
     return None
-  if rules is None:
-    rules = _axis_rules.rules
-  axis_name_counts = collections.Counter(array_dim_names)
-  dups = tuple(k for k, v in axis_name_counts.items()
-               if v > 1 and k is not None)
-  if dups:
-    raise ValueError(
-        f'Unsupported: Dimensions {dups} occur more than once in array names.')
-  if not isinstance(rules, (tuple, list)):
-    raise ValueError('Unknown axis rule specification type.')
-  # We assign mesh axes using a priority based ruleset over logical axis names.
-  result: List[Union[_UnassignedAxis, None, str, Tuple[str]]]
-  result = [_unassigned_axis] * len(array_dim_names)
-  for rule_model_name, rule_mesh_names in rules:
-    if rule_model_name in array_dim_names:
-      pos = array_dim_names.index(rule_model_name)
-      if (_mesh_assignment_free(rule_mesh_names, result) and
-          result[pos] == _unassigned_axis):
-        result[pos] = rule_mesh_names
   # We default to None - ie unsharded along the dimension.
   result = [None if x is _unassigned_axis else x for x in result]
   return pjit.PartitionSpec(*result)
@@ -191,9 +204,14 @@ def _global_mesh_defined() -> bool:
   return maps_env.physical_mesh.devices.shape != ()  # pylint: disable=g-explicit-bool-comparison
 
 
-def _with_sharding_constraint(
-    x: ArrayPytree,
-    axis_resources: PartitionSpecPytree):
+class RulesFallback(enum.Enum):
+  """How a sharding constraint should behave when no matching rule is found."""
+  AXIS_IS_UNSHARDED = 'axis_is_unsharded'
+  RAISE_ERROR = 'raise_error'
+  NO_CONSTRAINT = 'no_constraint'
+
+
+def _with_sharding_constraint(x: Array, axis_resources: pjit.PartitionSpec):
   """Wrapper for pjit with_sharding_constraint, no-op on cpu or outside pjit."""
   if jax.devices()[0].platform == 'cpu' or not _global_mesh_defined():
     return x
@@ -201,6 +219,26 @@ def _with_sharding_constraint(
     return pjit.with_sharding_constraint(x, axis_resources)
 
 
+def _with_sharding_constraint_one_fallback(
+    axis_resources: LogicalPartitionSpec,
+    x: Array,
+    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED):
+  """Either imposes a sharding constraint or applies fallback."""
+  mesh_axes = _logical_to_mesh_axes(axis_resources)
+  if mesh_axes is None:
+    return _with_sharding_constraint(x, None)
+
+  if fallback == RulesFallback.AXIS_IS_UNSHARDED:
+    mesh_axes = [None if x is _unassigned_axis else x for x in mesh_axes]
+  else:
+    if any(x is _unassigned_axis for x in mesh_axes):
+      if fallback == RulesFallback.RAISE_ERROR:
+        raise ValueError(f'Axis names {axis_resources} did not match a rule')
+      else:
+        return x
+  return _with_sharding_constraint(x, pjit.PartitionSpec(*mesh_axes))
+
+
 def _is_logical_spec(x):
   return x is None or (
       isinstance(x, tuple) and all(isinstance(e, str) or e is None for e in x))
@@ -208,15 +246,19 @@ def _is_logical_spec(x):
 
 def with_sharding_constraint(
     x: ArrayPytree,
-    logical_axis_resources: LogicalPartitionSpecPytree):
+    logical_axis_resources: LogicalPartitionSpecPytree,
+    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED):
   """Version of pjit's with_sharding_constraint that uses logical axis names."""
   # If no axis binding is set, this is a no-op.
   if not _axis_rules.rules or logical_axis_resources is None:
     return x
   # Translate logical names to mesh assignments.
-  axis_resources = jax.tree_map(logical_to_mesh_axes, logical_axis_resources,
-                                is_leaf=_is_logical_spec)
-  return _with_sharding_constraint(x, axis_resources)
+  return jax.tree_util.tree_map(
+      functools.partial(
+          _with_sharding_constraint_one_fallback, fallback=fallback),
+      logical_axis_resources,
+      x,
+      is_leaf=_is_logical_spec)
 
 
 # Annotated parameters and Module axis metadata handling.
@@ -313,8 +355,12 @@ class PartitionedVariable(flax.core.scope.Variable):
   and assignment.
   """
 
-  def __init__(self, scope, collection: str, name: str,
-               axes: Optional[Tuple[str, ...]] = None):
+  def __init__(self,
+               scope,
+               collection: str,
+               name: str,
+               axes: Optional[Tuple[str, ...]] = None,
+               fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED):
     """Initializes a partitioned variable.
 
     Args:
@@ -322,25 +368,27 @@ class PartitionedVariable(flax.core.scope.Variable):
       collection: The collection of the variable (e.g., "params").
       name: The name of the variable (e.g., "dense").
       axes: logical axes name of variable.
+      fallback: Fallback behavior if no matching rule is found.
     """
     self.scope = scope
     self.collection = collection
     self.name = name
     self.axes = axes
+    self.fallback = fallback
 
   @property
   def value(self):
     """Returns the value of this Variable."""
     value = self.scope.get_variable(self.collection, self.name)
     if self.axes is not None:
-      value = with_sharding_constraint(value, self.axes)
+      value = with_sharding_constraint(value, self.axes, fallback=self.fallback)
     return value
 
   @value.setter
   def value(self, value):
     """Updates the value of this Variable."""
     if self.axes is not None:
-      value = with_sharding_constraint(value, self.axes)
+      value = with_sharding_constraint(value, self.axes, fallback=self.fallback)
     self.scope.put_variable(self.collection, self.name, value)
 
 
@@ -350,7 +398,8 @@ def _core_variable_with_axes(
     name: str,
     init_fn: Callable[..., Any],
     *init_args,
-    axes: Tuple[str, ...] = ()):
+    axes: Tuple[str, ...] = (),
+    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED):
   """Variant of flax core variable scope call with sharding constraints."""
   scope.reserve(name)
   if not scope.has_variable(col, name):
@@ -358,9 +407,9 @@ def _core_variable_with_axes(
       raise flax.errors.ScopeVariableNotFoundError(name, col, scope.path_text)
     init_value = init_fn(*init_args)
     if axes:
-      init_value = with_sharding_constraint(init_value, axes)
+      init_value = with_sharding_constraint(init_value, axes, fallback=fallback)
     scope.put_variable(col, name, init_value)
-  return PartitionedVariable(scope, col, name, axes)
+  return PartitionedVariable(scope, col, name, axes, fallback)
 
 
 def variable_with_axes(
@@ -369,7 +418,8 @@ def variable_with_axes(
     init_fn,
     *init_args,
     axes: Tuple[str, ...] = (),
-    module: Optional[nn.Module] = None):
+    module: Optional[nn.Module] = None,
+    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED):
   """Declares and returns a variable with logical axes in the current Module.
 
   See :mod:`flax.linen.module.variable` for original docstring.
@@ -384,6 +434,7 @@ def variable_with_axes(
     axes: A tuple of axis names, must match the rank of the variable array.
     module: Use an explicit module instead of deriving the most recent from
       dynamic module context.
+    fallback: How sharding should behave if there is no rule covering some axis.
 
   Returns:
     A flax `PartitionedVariable` object referencing the initialized variable
@@ -398,7 +449,13 @@ def variable_with_axes(
     module = nn.module._context.module_stack[-1]  # pylint: disable=protected-access
     assert module is not None
   module_var = _core_variable_with_axes(
-      module.scope, collection, name, init_fn, *init_args, axes=axes)
+      module.scope,
+      collection,
+      name,
+      init_fn,
+      *init_args,
+      axes=axes,
+      fallback=fallback)
   if axes:
     # record logical axis constraint for global axis metadata
     module.sow(
@@ -422,7 +479,7 @@ def get_axis_names(axes_metadata):
   def leaf_rewrite(x):
     return None if x is None else pjit.PartitionSpec(*x)
   def rewrite(tree):
-    return jax.tree_map(leaf_rewrite, tree, is_leaf=_is_logical_spec)
+    return jax.tree_util.tree_map(leaf_rewrite, tree, is_leaf=_is_logical_spec)
   axes_metadata = unfreeze(axes_metadata)  # pytype: disable=wrong-arg-types
   flat_dict = {
       re.sub(r'_axes$', '', '/'.join(k)): rewrite(v.names)
@@ -439,7 +496,7 @@ def get_axis_names(axes_metadata):
 def _tree_map_axes(fn, tree):
   """Only map over AxisMetadata leaves in pytree - identity for other leaves."""
   safe_fn = lambda x: fn(x) if isinstance(x, AxisMetadata) else x
-  return jax.tree_map(
+  return jax.tree_util.tree_map(
       safe_fn, tree, is_leaf=lambda x: isinstance(x, AxisMetadata))
 
 
@@ -480,7 +537,8 @@ def _add_axis_to_metadata(fn, axis_pos, axis_name, axis_col='params_axes'):
     return tuple(names)
 
   def insert_fn(x):
-    new_names = jax.tree_map(insert_fn_leaf, x.names, is_leaf=_is_logical_spec)
+    new_names = jax.tree_util.tree_map(insert_fn_leaf, x.names,
+                                       is_leaf=_is_logical_spec)
     return x.replace(names=new_names)
 
   def remove_fn_leaf(names):
@@ -494,7 +552,8 @@ def _add_axis_to_metadata(fn, axis_pos, axis_name, axis_col='params_axes'):
     return tuple(names)
 
   def remove_fn(x):
-    new_names = jax.tree_map(remove_fn_leaf, x.names, is_leaf=_is_logical_spec)
+    new_names = jax.tree_util.tree_map(remove_fn_leaf, x.names,
+                                       is_leaf=_is_logical_spec)
     return x.replace(names=new_names)
 
   return nn.transforms.map_variables(
@@ -518,6 +577,7 @@ def scan_with_axes(
     out_axes=0,
     length: Optional[int] = None,
     reverse: bool = False,
+    unroll: int = 1,
     axis_name: str = 'layers',
     axes_collections: Tuple[str, ...] = ('params',),
     data_transform: Optional[Callable[..., Any]] = None,
@@ -541,6 +601,7 @@ def scan_with_axes(
       out_axes=out_axes,
       length=length,
       reverse=reverse,
+      unroll=unroll,
       data_transform=data_transform,
       methods=methods)
 
@@ -565,6 +626,7 @@ def vmap_with_axes(target: flax.linen.transforms.Target,
                    axis_size: Optional[int] = None,
                    axis_name: Optional[str] = None,
                    partitioning_axis_names: Mapping[str, str] = {},
+                   spmd_axis_name: Optional[str] = None,
                    methods=None) -> flax.linen.transforms.Target:
   """Wrapped version of nn.vmap that handles logical axis metadata."""
 
@@ -583,6 +645,7 @@ def vmap_with_axes(target: flax.linen.transforms.Target,
       out_axes=out_axes,
       axis_size=axis_size,
       axis_name=axis_name,
+      spmd_axis_name=spmd_axis_name,
       methods=methods)
 
   for collection_name, axis in variable_axes.items():
