@@ -38,10 +38,13 @@ from flax.linen import module as linen_module
 from flax.linen.module import Module
 from flax.linen.module import Variable
 from flax.linen.module import wrap_method_once
+from flax.linen.module import _get_unbound_fn
+from flax.linen.module import _derive_profiling_name
 import jax
 
 traceback_util.register_exclusion(__file__)
 
+# pylint: disable=protected-access
 
 # Utils
 # -----------------------------------------------------------------------------
@@ -78,12 +81,12 @@ def _memoize_by_id(fn, refs):
     nonlocal refs
     if isinstance(x, (VariablePlaceholder, InstancePlaceholder)):
       x_id = x.id
+    elif isinstance(x, (Variable, Module)):
+      x_id = x._id
     else:
-      x_id = id(x)
+      return fn(x)
     if x_id not in refs:
       refs[x_id] = fn(x)
-    else:
-      pass
     return refs[x_id]
   return wrapped_fn
 
@@ -123,25 +126,25 @@ def get_module_scopes(module, args=None, kwargs=None):
     nonlocal scopes
     if isinstance(x, Variable) and isinstance(x.scope, Scope):
       scopes.append(x.scope)
-      return VariablePlaceholder(x.collection, x.name, id(x))
+      return VariablePlaceholder(x.collection, x.name, x._id)
     elif isinstance(x, Module) and isinstance(x.scope, Scope):
-      x._try_setup(shallow=True)  # pylint: disable=protected-access
+      x._try_setup(shallow=True)
       scopes.append(x.scope)
       attrs = {
           f.name: getattr(x, f.name)
           for f in dataclasses.fields(x)
           if f.name != 'parent' and f.init
       }
-      attrs = jax.tree_map(get_arg_scope, attrs)
-      return InstancePlaceholder(x.__class__, attrs, id(x))
+      attrs = jax.tree_util.tree_map(get_arg_scope, attrs)
+      return InstancePlaceholder(x.__class__, attrs, x._id)
     return x
-  new_args, new_kwargs = jax.tree_map(get_arg_scope, (args, kwargs))
+  new_args, new_kwargs = jax.tree_util.tree_map(get_arg_scope, (args, kwargs))
 
   # Gather scopes in Variables and Submodules passed as Module attributes.
   @functools.partial(_memoize_by_id, refs=refs)
   def get_scopes(module):
     nonlocal scopes
-    module._try_setup(shallow=True)  # pylint: disable=protected-access
+    module._try_setup(shallow=True)
     def get_scopes_inner(x):
       nonlocal scopes
       if isinstance(x, Module) and isinstance(x.scope, Scope):
@@ -154,7 +157,7 @@ def get_module_scopes(module, args=None, kwargs=None):
         for f in dataclasses.fields(module)
         if f.name != 'parent' and f.init
     }
-    jax.tree_map(get_scopes_inner, attrs)
+    jax.tree_util.tree_map(get_scopes_inner, attrs)
     scopes.append(module.scope)
   get_scopes(module)
   return scopes, new_args, new_kwargs
@@ -200,14 +203,14 @@ def set_module_scopes(module, args, kwargs, scopes):
     elif isinstance(x, InstancePlaceholder):
       instance_scope = scopes[idx]
       idx += 1
-      instance_attrs = jax.tree_map(set_arg_scope, x.attrs)
+      instance_attrs = jax.tree_util.tree_map(set_arg_scope, x.attrs)
       return x.cls(parent=instance_scope, **instance_attrs)
     return x
 
   def is_placeholder(x):
     return isinstance(x, (VariablePlaceholder, InstancePlaceholder))
 
-  new_args, new_kwargs = jax.tree_map(
+  new_args, new_kwargs = jax.tree_util.tree_map(
       set_arg_scope, (args, kwargs), is_leaf=is_placeholder)
 
   # set scopes in Variables and Submodules passed as Module attributes
@@ -232,7 +235,7 @@ def set_module_scopes(module, args, kwargs, scopes):
         for f in dataclasses.fields(module)
         if f.name != 'parent' and f.init
     }
-    new_attrs = jax.tree_map(set_scopes_inner, attrs)
+    new_attrs = jax.tree_util.tree_map(set_scopes_inner, attrs)
     new_module = module.clone(parent=scopes[idx], **new_attrs)
     idx += 1
     return new_module
@@ -244,7 +247,7 @@ def set_module_scopes(module, args, kwargs, scopes):
 def _test_transformed_return_values(tree, method_name):
   """Tests whether the return value contains any Modules or Variables."""
   impure = any(map(lambda x: isinstance(x, (Module, Variable)),
-                   jax.tree_leaves(tree)))
+                   jax.tree_util.tree_leaves(tree)))
   if impure:
     raise errors.TransformedMethodReturnValueError(method_name)
 
@@ -302,9 +305,9 @@ def module_class_lift_transform(
         # we reference module_class, not self.__class__ to avoid infinite loop
         cloned = module_class(parent=None, **attrs)
         cloned, args, kwargs = set_module_scopes(cloned, args, kwargs, scopes)
-        object.__setattr__(cloned, '_state', state.export())  # pylint: disable=protected-access
+        object.__setattr__(cloned, '_state', state.export())
         res = fn(cloned, *args, **kwargs)
-        self._state.reimport(cloned._state)  # pylint: disable=protected-access
+        self._state.reimport(cloned._state)
         _test_transformed_return_values(res, fn_name)
         return res
       # here we apply the given lifting transform to the scope-ingesting fn
@@ -350,9 +353,9 @@ def decorator_lift_transform(transform, class_fn, *trafo_args,
       if not multi_scope:
         scopes = [scopes]
       cloned, args, kwargs = set_module_scopes(self, args, kwargs, scopes)
-      object.__setattr__(cloned, '_state', state.export())  # pylint: disable=protected-access
+      object.__setattr__(cloned, '_state', state.export())
       res = prewrapped_fn(cloned, *args, **kwargs)
-      self._state.reimport(cloned._state)  # pylint: disable=protected-access
+      self._state.reimport(cloned._state)
       _test_transformed_return_values(res, getattr(class_fn, '__name__', None))
       return res
     core_fns = [functools.partial(core_fn, prewrapped_fn, class_fn)
@@ -419,6 +422,8 @@ def lift_direct_transform(transform: Callable[..., Any],
           ' That is function that takes a Module instance as its first arg.')
     elif not callable(target):
       raise ValueError('transform target must be callable')
+  # normalize self.foo bound methods to class.foo unbound methods.
+  targets = tuple(_get_unbound_fn(target) for target in targets)
   aug_transform = lambda *fns: functools.partial(transform, *fns)
   return decorator_lift_transform(
       aug_transform, targets, multi_scope=multi_scope)(mdl, *args, **kwargs)
@@ -671,6 +676,7 @@ def scan(target: Target,
          in_axes=0, out_axes=0,
          length: Optional[int] = None,
          reverse: bool = False,
+         unroll: int = 1,
          data_transform: Optional[Callable[..., Any]] = None,
          methods=None) -> Target:
   """A lifted version of ``jax.lax.scan``.
@@ -763,6 +769,8 @@ def scan(target: Target,
     length: Specifies the number of loop iterations. This only needs to be
       specified if it cannot be derivied from the scan arguments.
     reverse: If true, scan from end to start in reverse order.
+    unroll: how many scan iterations to unroll within a single
+      iteration of a loop (default: 1).
     data_transform: optional function to transform raw functional-core variable
       and rng groups inside lifted scan body_fn, intended for inline SPMD
       annotations.
@@ -782,6 +790,7 @@ def scan(target: Target,
       in_axes=in_axes, out_axes=out_axes,
       length=length,
       reverse=reverse,
+      unroll=unroll,
       data_transform=data_transform,
       methods=methods)
 
@@ -803,7 +812,7 @@ def map_variables(
       @nn.compact
       def __call__(self, x):
         def sign(x):
-          return jax.tree_map(jnp.sign, x)
+          return jax.tree_util.tree_map(jnp.sign, x)
         MapDense = nn.map_variables(nn.Dense, "params", sign, init=True)
         return MapDense(4)(x)
 
@@ -939,7 +948,8 @@ def jvp(
       @nn.compact
       def __call__(self, x):
         scale = LearnScale()
-        vars_t = jax.tree_map(jnp.ones_like, scale.variables.get('params', {}))
+        vars_t = jax.tree_util.tree_map(jnp.ones_like,
+                                        scale.variables.get('params', {}))
         _, out_t = nn.jvp(
             lambda mdl, x: mdl(x), scale, (x,), (jnp.zeros_like(x),),
             variable_tangents={'params': vars_t})
@@ -952,7 +962,7 @@ def jvp(
       return p * x
 
     def f(scope, x):
-      vars_t = jax.tree_map(jnp.ones_like, scope.variables().get('params', {}))
+      vars_t = jax.tree_util.tree_map(jnp.ones_like, scope.variables().get('params', {}))
       x, out_t = lift.jvp(
           learn_scale, scope, (x,), (jnp.zeros_like(x),),
           variable_tangents={'params': vars_t})
@@ -1064,10 +1074,10 @@ def _cond_wrapper(t_fn, f_fn, scope, pred, *ops, variables, rngs):
 
 
 def cond(
-    pred: Any, 
+    pred: Any,
     true_fun: Callable[..., C], false_fun: Callable[..., C],
     mdl: Module, *operands,
-    variables: lift.CollectionFilter = True, 
+    variables: lift.CollectionFilter = True,
     rngs: lift.PRNGSequenceFilter = True) -> C:
   """Lifted version of ``jax.lax.cond``.
 
@@ -1078,7 +1088,7 @@ def cond(
   Note that this constraint is violated when
   creating variables or submodules in only one branch.
   Because initializing variables in just one branch
-  causes the paramater structure to be different.
+  causes the parameter structure to be different.
 
   Example::
 
@@ -1094,15 +1104,15 @@ def cond(
           mdl.variable('state', 'false_count').value += 1
           return -nn.Dense(2, name='dense')(x)
         return nn.cond(pred, true_fn, false_fn, self, x)
-  
-  
+
+
   Args:
     pred: determines if true_fun or false_fun is evaluated.
     true_fun: The function evalauted when ``pred`` is `True`.
-      The signature is (Scope, *operands) -> T.
+      The signature is (module, *operands) -> T.
     false_fun: The function evalauted when ``pred`` is `False`.
-      The signature is (Scope, *operands) -> T.
-    scope: A Scope or Pytree of scopes to pass 
+      The signature is (module, *operands) -> T.
+    mdl: A Module target to pass.
     *operands: The arguments passed to ``true_fun`` and ``false_fun``
     variables: The variable collections passed to the conditional
       branches (default: all)
@@ -1115,6 +1125,89 @@ def cond(
       pred, *operands,
       variables=variables, rngs=rngs)
 
+def _switch_wrapper(*args, variables, rngs, n_branches):
+  # first n_branches arguments are branches.
+  # then scope, index, and the rest are *operands
+  branches = args[:n_branches]
+  scope, index, *operands = args[n_branches:]
+  return lift.switch(index, branches, scope, *operands, variables=variables, rngs=rngs)
+
+def switch(
+    index: Any,
+    branches: Sequence[Callable[..., C]],
+    mdl: Module, *operands,
+    variables: lift.CollectionFilter = True,
+    rngs: lift.PRNGSequenceFilter = True) -> C:
+  """Lifted version of ``jax.lax.switch``.
+
+  The returned values from ``branches``
+  must have the same Pytree structure, shapes, and dtypes.
+  The variables created or updated inside the
+  branches must also have the same structure.
+  Note that this constraint is violated when
+  creating variables or submodules in only one branch.
+  Because initializing variables in just one branch
+  causes the parameter structure to be different.
+
+  Example::
+
+    class SwitchExample(nn.Module):
+      @nn.compact
+      def __call__(self, x, index):
+        self.variable('state', 'a_count', lambda: 0)
+        self.variable('state', 'b_count', lambda: 0)
+        self.variable('state', 'c_count', lambda: 0)
+        def a_fn(mdl, x):
+          mdl.variable('state', 'a_count').value += 1
+          return nn.Dense(2, name='dense')(x)
+        def b_fn(mdl, x):
+          mdl.variable('state', 'b_count').value += 1
+          return -nn.Dense(2, name='dense')(x)
+        def c_fn(mdl, x):
+          mdl.variable('state', 'c_count').value += 1
+          return nn.Dense(2, name='dense')(x)
+        return nn.switch(index, [a_fn, b_fn, c_fn], self, x)
+
+  If you want to have a different parameter structure for each branch
+  you should run all branches on initialization before calling switch::
+
+    class MultiHeadSwitchExample(nn.Module):
+      def setup(self) -> None:
+        self.heads = [
+          nn.Sequential([nn.Dense(10), nn.Dense(7), nn.Dense(5)]),
+          nn.Sequential([nn.Dense(11), nn.Dense(5)]),
+          nn.Dense(5),
+        ]
+      
+      @nn.compact
+      def __call__(self, x, index):
+        def head_fn(i):
+          return lambda mdl, x: mdl.heads[i](x)
+        branches = [head_fn(i) for i in range(len(self.heads))]
+
+        # run all branches on init
+        if self.is_mutable_collection('params'):
+          for branch in branches:
+            _ = branch(self, x)
+          
+        return nn.switch(index, branches, self, x)
+
+  Args:
+    index: Integer scalar type, indicating which branch function to apply.
+    branches: Sequence of functions to be applied based on index.
+      The signature of each function is (module, *operands) -> T.
+    mdl: A Module target to pass.
+    *operands: The arguments passed to the branches.
+    variables: The variable collections passed to the conditional
+      branches (default: all)
+    rngs: The PRNG sequences passed to the conditionals (default: all)
+  Returns:
+    The result of the evaluated branch.
+  """
+  return lift_direct_transform(
+      _switch_wrapper, tuple(branches), mdl,
+      index, *operands,
+      variables=variables, rngs=rngs, n_branches=len(branches))
 
 # a version of lift.custom_vjp with a single scope function
 # this avoids having to lift multiple functions in
@@ -1165,7 +1258,7 @@ def custom_vjp(fn: Callable[..., Any],
 
         def bwd(vjp_fn, y_t):
           input_t, params_t = vjp_fn(y_t)
-          params_t = jax.tree_map(jnp.sign, params_t)
+          params_t = jax.tree_util.tree_map(jnp.sign, params_t)
           return input_t, params_t
 
         sign_grad = nn.custom_vjp(
@@ -1204,10 +1297,12 @@ def custom_vjp(fn: Callable[..., Any],
       multi_scope=False)
 
 
-# Special case of decorator_lift_transform to handle named calls for profiling.
 def named_call(class_fn, force=True):
   """Labels a method for labelled traces in profiles.
 
+  Note that it is better to use the `jax.named_scope` context manager directly
+  to add names to JAX's metadata name stack.
+
   Args:
     class_fn: The class method to label.
     force: If True, the named_call transform is applied even if it is globally
@@ -1215,43 +1310,12 @@ def named_call(class_fn, force=True):
   Returns:
     A wrapped version of ``class_fn`` that is labeled.
   """
-  if (hasattr(jax.config, 'jax_experimental_name_stack') and
-      jax.config.jax_experimental_name_stack):
-    # Use JAX's improved dynamic name-stack named_call.
-    # No transform boundary needed!
-    @functools.wraps(class_fn)
-    def wrapped_fn(self, *args, **kwargs):
-      fn_name = class_fn.__name__
-      method_suffix = f'.{fn_name}' if fn_name != '__call__' else ''
-      module_name = self.name or self.__class__.__name__
-      full_name = f'{module_name}{method_suffix}'
-      return jax.named_call(class_fn, name=full_name)(self, *args, **kwargs)
-  else:
-    # Use JAX's old purely-functional call-based named_call.
-    # Due to the ordering of method decorators, we must wrap the class_fn
-    # with the module state management wrapper first to maintain Module state
-    # correctly.
-    prewrapped_fn = wrap_method_once(class_fn)
-    @functools.wraps(prewrapped_fn)
-    def wrapped_fn(self, *args, **kwargs):
-      if ((not force and not linen_module._use_named_call)  # pylint: disable=protected-access
-          or self._state.in_setup):  # pylint: disable=protected-access
-        return prewrapped_fn(self, *args, **kwargs)
-      fn_name = class_fn.__name__
-      method_suffix = f'.{fn_name}' if fn_name != '__call__' else ''
-      module_name = self.name or self.__class__.__name__
-      full_name = f'{module_name}{method_suffix}'
-      # make a scope-function to transform
-      def core_fn(scopes, *args, **kwargs):
-        cloned, args, kwargs = set_module_scopes(self, args, kwargs, scopes)
-        object.__setattr__(cloned, '_state', self._state.export())  # pylint: disable=protected-access
-        res = prewrapped_fn(cloned, *args, **kwargs)
-        self._state.reimport(cloned._state)  # pylint: disable=protected-access
-        _test_transformed_return_values(res, fn_name)
-        return res
-      # here we apply the given lifting transform to the scope-ingesting fn
-      trafo_fn = lift.named_call(core_fn, full_name)
-      module_scopes, args, kwargs = get_module_scopes(self, args, kwargs)
-      return trafo_fn(module_scopes, *args, **kwargs)
-
+  # We use JAX's dynamic name-stack named_call. No transform boundary needed!
+  @functools.wraps(class_fn)
+  def wrapped_fn(self, *args, **kwargs):
+    if ((not force and not linen_module._use_named_call)  # pylint: disable=protected-access
+        or self._state.in_setup):  # pylint: disable=protected-access
+      return class_fn(self, *args, **kwargs)
+    full_name = _derive_profiling_name(self, class_fn)
+    return jax.named_call(class_fn, name=full_name)(self, *args, **kwargs)
   return wrapped_fn
